             .title        "LS161 CHR Reader Plugin"

;07/23/10 
;Written by BootGod
;Version 1.0

;Somewhat generic CNROM-like plugin that can be used for greater than 32KB of CHR
;for carts using a LS161 D0-D3 for bank selection. Cannot be used for official CNROM
;carts that use special features like diodes and/or special CE masks.
             
             ;vectors for standard system calls

send_byte:   .equ 0200h
baton:       .equ 0203h
chk_vram:    .equ 0206h
chk_wram:    .equ 0209h
wr_ppu:      .equ 020ch
read_byte:   .equ 020fh
init_crc:    .equ 0212h
do_crc:      .equ 0215h
finish_crc:  .equ 0218h

crc0:        .equ 0080h
crc1:        .equ 0081h
crc2:        .equ 0082h
crc3:        .equ 0083h

addr1:       .equ 00e0h
addr1_lo:    .equ 00e0h
addr1_hi:    .equ 00e1h
addr2:       .equ 00e2h
addr2_lo:    .equ 00e2h
addr2_hi:    .equ 00e3h

curbank:     .equ 00e4h
reg:         .equ 00e5h
banksleft:   .equ 00e6h
tx:          .equ 00e7
numpages_lo  .equ 00e8h
numpages_hi  .equ 00e9h
bank0crc:    .equ 00eah
temp_crc:    .equ 00eeh


;plugin header defines
p_ver: = 1	;plugin version
p_read: = 1	;plugin is for downloading (dumping)
p_write: = 2	;plugin is for uploading (RAM carts / WRAM etc)
p_other: = 3	;plugin is for some other purpose (i.e. BankWatch)
p_var1: = 4	;plugin uses user variable #1
p_var2: = 8	;plugin uses user variable #2
p_var3: = 16	;plugin uses user variable #3
p_var4: = 32	;plugin uses user variable #4

;plugin header data

             .org 0380h
             
             ;plugin description
             .db "LS161 D0-D3 CHR carts"
             .fill 03e0h-*,00h

             ;plugin creator(s)
             .db "BootGod"
             .fill 03f8h-*,00h

             ;PRG/CHR/WRAM min/max sizes supported
             .db 16 >> 4,32 >> 4
             .db 8 >> 3,128 >> 3
             .db 0,0

             ;plugin version and flags
             .db p_ver
             .db p_read

;plugin code start here
             
;check mirroring (x = mirroring)
             ldx #00h		;Horizontal
             lda #020h
             jsr wr_ppu
             lda #055h
             sta 2007h
             lda #0aah
             sta 2007h
             lda #024h
             jsr wr_ppu
             lda 2007h
             lda 2007h
             cmp #055h
             bne got_mir
             lda 2007h
             cmp #0aah
             bne got_mir
             lda #020h
             jsr wr_ppu
             lda #0aah
             sta 2007h
             lda #024h
             jsr wr_ppu
             lda 2007h
             lda 2007h
             cmp #0aah
             bne got_mir
             inx		;Vertical
got_mir:     txa
             jsr send_byte	;send mirroring
             
;determine PRG ROM size (16K or 32K)
             lda #000h
             sta addr1_lo
             sta addr2_lo
             lda #080h
             sta addr1_hi
             lda #0c0h
             sta addr2_hi	;set addresses to $8000 and $C000
             
             ldx #040h
             ldy #0		;set up to check 16K

chk_prg:     lda (addr1),y
             cmp (addr2),y
             bne is32k
             iny
             bne chk_prg
             inc addr1_hi
             inc addr2_hi
             dex
             bne chk_prg

is16k:       lda #040h		;40h * 256bytes = 16K
             tax
             bne send_prg

is32k:       lda #080h		;80h * 256bytes = 32K
             tax

send_prg:    jsr send_byte
             lda #000h
             jsr send_byte
             lda #001h
             jsr send_byte	;send PRG header

             ldy #0
             sty addr1_lo
             lda #080h
             sta addr1_hi	;set address to $8000, X = num blocks to send

dump_prg:    lda (addr1),y
             jsr send_byte
             iny
             bne dump_prg
             inc addr1_hi
             dex
             bne dump_prg
             
;determine CHR ROM size (16K - 128K)
             lda #0
             jsr wr_bank	;set to CHR bank 0
             jsr crc_cbank

             ldx #3
ld_ccrc:     lda temp_crc,x	;save for comparision
             sta bank0crc,x
             dex
             bpl ld_ccrc
             
;start out crc check at bank 1
             lda #1
             sta curbank

             lda #0		;pages init at 8K
             sta numpages_hi
             lda #020h
             sta numpages_lo

next_ccrc:   lda curbank
             jsr wr_bank	;set CHR bank num
             jsr crc_cbank
             
;compare crc's of bank 0 and curbank
             ldx #3
chk_ccrc:    lda bank0crc,x
             cmp temp_crc,x
             bne crc_cfail
             dex
             bpl chk_ccrc

             jmp got_chr

crc_cfail:   lda curbank	;check to see if we've hit the max bank num (16)
             cmp #16
             beq got_chr	;stop detection if we have

             asl curbank
             asl numpages_lo
             rol numpages_hi
             
             jmp next_ccrc
             
got_chr:     lda curbank	;setup bank counters for dumping
             sta banksleft
             lda #0
             sta curbank

             lda numpages_lo	;send CHR header
             jsr send_byte
             lda numpages_hi
             jsr send_byte
             lda #002h
             jsr send_byte
             
next_chr:    lda curbank
             jsr wr_bank
             inc curbank
             lda #0
             jsr wr_ppu
             lda 2007h

             ldx #32
             ldy #0

dump_chr:    lda 2007h
             jsr send_byte
             iny
             bne dump_chr
             dex
             bne dump_chr
             dec banksleft
             bne next_chr
             
done:        lda #000h		;send end block
             jsr send_byte
             lda #000h
             jsr send_byte
             lda #000h
             jsr send_byte

             rts		;return to BIOS
             

;sub for setting CHR bank (compensates for bus conflicts)

wr_bank:     sta reg		;store value to be written to reg

             lda #0
             sta addr1_lo
             lda #080h
             sta addr1_hi	;set address to $8000

             ldx #080h		;setup to search all 32K of PRG
             ldy #0

wrb_loop:    lda (addr1),y
             and #0fh		;mask it to just the bits we need to match
             cmp reg		;check if current byte matches what we want to write
             beq found_reg
             iny
             bne wrb_loop
             inc addr1_hi
             bne wrb_loop

no_reg:      lda reg		;didn't find place to write, try it anyways
             sta 08000h
             jmp wrb_return

found_reg:   lda (addr1),y	;found value equal to reg, write it.
             sta (addr1),y

wrb_return:  rts

;CRC routine for 8K CHR bank @ $0000, stores crc in temp_crc
crc_cbank:   lda #0
             jsr wr_ppu
             lda 2007h		;set PPU address to $0000

             jsr init_crc
             ldx #32
             ldy #0		;init crc for 8K of data (32 pages = 8K)

crc_cloop:   lda 2007h		;crc generation loop
             stx tx
             jsr do_crc
             ldx tx
             iny
             bne crc_cloop
             dex
             bne crc_cloop
             
             jsr finish_crc	;crc done

             ldx #3
crc_csave:   lda crc0,x		;save crc in temp_crc
             sta temp_crc,x
             dex
             bpl crc_csave

             rts

             .fill 0800h-*,0ffh   ;fill rest to get 1K of data

             .end
